/**
 * Juliac
 * Copyright (C) 2009-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Christophe Munilla
 */
package org.objectweb.fractal.juliac.osgi.concierge;

import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.jar.JarFile;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.FrameworkEvent;

import ch.ethz.iks.concierge.framework.Framework;

/**
 * Concierge OSGi Framework Launcher and extension
 */
public class Concierge
{
    private static final String FRAMEWORK_CLASS_NAME =
            "ch.ethz.iks.concierge.framework.Framework";
    
    private static final String BUNDLECONTEXT_CLASS_NAME = FRAMEWORK_CLASS_NAME
            + ".BundleContextImpl";
    
    private static final String SYSTEMBUNDLE_CLASS_NAME = FRAMEWORK_CLASS_NAME
            + ".SystemBundle";
    
    private static Concierge instance; 
    
    /**
     * Logger
     */
    private static final Logger LOGGER = Logger.getLogger(
            Concierge.class.getCanonicalName());
    
    /** Concierge Framework's inner-classes */
    private Class<?>[] classiz = null;
    
    /** Inner-class BundleContextImpl */
    private Class<?> bundleContextClass = null;
    
    /** Inner-class BundleContextImpl constructor */
    private Constructor<?> bundleContextConstructor = null;
    
    /** Concierge Framework's SystemBundle instance */
    private Bundle systemBundle = null;
    
    /**
     * Concierge-R3 class implementation of the org.osgi.framework.Bundle interface
     */
    private Class<?> bundleClass = null;

    /**
     * Create the Concierge framework instance if it doesn't exist
     * and return it
     * 
     * @return
     *          the Concierge framework instance
     */
    public static final Concierge getInstance(File cache)
    {
        if(instance == null)
        {
            instance = new Concierge(cache);
        }
        return instance;
    } 
    
    /**
     * Return the Concierge framework instance
     * 
     * @return
     *          the Concierge framework instance
     */
    public static final Concierge getInstance()
    {
        return instance;
    }
    
    /**
     * The Concierge Framework Constructor
     */
    @SuppressWarnings({ "rawtypes", "unchecked" }) 
    private Concierge(File cache)
    {
        if (systemBundle != null)
        {
            invokeFrameworkMethod("shutdown", 
                    new Class<?>[] { boolean.class }, 
                    new Object[] { false });
        }
        setSystemPackagesFieldValue(System.getProperty(
                "org.osgi.framework.system.packages",""));
        try
        {
            // bundles' storage
            System.setProperty("ch.ethz.iks.concierge.basedir", cache.getAbsolutePath());
            try
            {
                System.setProperty("ch.ethz.iks.concierge.jars",
                    new File(cache.getAbsolutePath()+File.separator + "data").toURI(
                            ).toURL().toExternalForm());
            } catch (MalformedURLException e)
            {
                LOGGER.log(Level.CONFIG,e.getMessage(),e);
            }
            System.setProperty("ch.ethz.iks.concierge.storage", 
                    cache.getAbsolutePath() + File.separator + "storage");
            
            // start the framework
            setFrameworkFieldValue("frameworkStartupShutdown", true);
            setFrameworkFieldValue("frameworkThread", Thread.currentThread());
            setFrameworkFieldValue("properties", System.getProperties());
            
            Properties properties = (Properties) getFrameworkFieldValue("properties");
            System.out.println("-----------------------------------------");
            System.out.println("  Concierge OSGi "
                    + getFrameworkFieldValue("FRAMEWORK_VERSION") + " on "
                    + properties.get("os.name") + " "
                    + properties.get("os.version") + " starting ...");
            System.out.println("----------------------------------------");
            
            long time = 0;
            time = System.currentTimeMillis();
            
            setFrameworkFieldValue("PROFILE", "");            
            String propertyFilePath = System.getProperty("xargs");            
            if (propertyFilePath != null)
            {
                final File propertyFile = new File(propertyFilePath);
                if (propertyFile.exists())
                {
                    invokeFrameworkMethod("processXargsFile", new Class<?>[]
                    { File.class }, new Object[]
                    { propertyFile });
                }
            }
            invokeFrameworkMethod("initialize", (Class<?>[]) null,
                    (Object[]) null);
            invokeFrameworkMethod("launch", (Class<?>[]) null, (Object[]) null);
            setFrameworkFieldValue("restart", false);
            findBundleContextContructor();
            systemBundle = (Bundle) getFrameworkFieldValue("systemBundle");
            
            String install = properties.getProperty("osgi.auto.install.1","");
            String start = properties.getProperty("osgi.auto.start.1","");
            
            final String[] str = { install, start };
            if(install != null || start != null)
            {
                for (int i = 0; i < 2; i++)
                {
                    if (str[i] != null)
                    {
                        StringTokenizer tokenizer = new StringTokenizer(str[i], " ");
                        while (tokenizer.hasMoreTokens())
                        {
                            final String location = tokenizer.nextToken();
                            autoInstall(location,1,false);
                        }
                    }
                }
            }
            // set startlevel and start all bundles that are marked to be
            // started up to the intended startlevel
            List bundles = (List) getFrameworkFieldValue("bundles");
            invokeSystemBundleMethod("setLevel", new Class<?>[]
            { Bundle[].class, int.class, boolean.class }, new Object[]
            { bundles.toArray(new Bundle[bundles.size()]), 1, false });
            
            setFrameworkFieldValue("frameworkStartupShutdown", false);
            setFrameworkFieldValue("DECOMPRESS_EMBEDDED",true);
            
            invokeFrameworkMethod("storeProfile", (Class<?>[]) null, (Object[]) null);
            final float timediff = (System.currentTimeMillis() - time) / (float) 1000.00;
            
            System.out.println("----------------------------------------");
            System.out.println(" Framework started in " + timediff + " seconds.");
            System.out.println("----------------------------------------");
            System.out.flush();
            
            setSystemBundleFieldValue("state", Bundle.ACTIVE);
            invokeFrameworkMethod("notifyFrameworkListeners", new Class<?>[]
            { int.class, Bundle.class, Throwable.class }, new Object[]
            { FrameworkEvent.STARTED, systemBundle, null });

        } catch (Throwable e)
        {
            LOGGER.log(Level.CONFIG,e.getMessage(),e);
            
        }
    }

    /**
     * 
     * @param location
     * @param level
     * @param persistently
     * @return
     */
    private Object autoInstall(String location,
            int level,boolean persistently)
    {
        try
        {
            Object bundle = invokeFrameworkMethod( "installNewBundle",
              new Class<?>[] { String.class }, 
              new Object[] { location });
            Field levelField = bundle.getClass().getDeclaredField("currentStartlevel");
            levelField.setAccessible(true);
            levelField.set(bundle, level);
            if (persistently)
            {
                Field persist = bundle.getClass().getDeclaredField("persistently");
                persist.setAccessible(true);
                persist.set(bundle, true);
            }
            return bundle;
        } catch (Exception e)
        {
            LOGGER.log(Level.CONFIG,e.getMessage(),e);
        }
        return null;
    }
    
    /** 
     * Retrieve an inner-class of the framework  
     * @param className
     *            the name of the class we look for
     */
    private Class<?> getFrameworkInnerClass(String className)
    {
        if (classiz == null)
        {
            classiz = Framework.class.getDeclaredClasses();
        }
        int i = 0;
        for (; i < classiz.length; i++)
        {
            if (classiz[i].getCanonicalName().equals(className))
            {
                break;
            }
        }
        if (i < classiz.length)
        {
            return classiz[i];
        }
        return null;
    }

    /**
     * Retrieve the no-args constructor of the BundleContextImpl
     */
    private void findBundleContextContructor()
    {
        bundleContextClass = getFrameworkInnerClass(BUNDLECONTEXT_CLASS_NAME);
        Method constructorMethod;
        try
        {
            constructorMethod = Class.class.getDeclaredMethod("getDeclaredConstructors0",
                    new Class<?>[] { boolean.class });
            constructorMethod.setAccessible(true);
            Constructor<?>[] constructors = (Constructor<?>[]) 
            constructorMethod.invoke(bundleContextClass, new Object[]{ false });
            
            bundleContextConstructor = constructors[0];
            bundleContextConstructor.setAccessible(true);
            
        } catch (Throwable e)
        {
            LOGGER.log(Level.CONFIG,e.getMessage(),e);
            
        }
    }

    /**
     * Return the value of a static field of the Framework
     * 
     * @param fieldName
     *            the name of the field
     */
    private Object getFrameworkFieldValue(String fieldName)
    {
        Object fieldObject = null;
        Field field;
        try
        {
            field = Framework.class.getDeclaredField(fieldName);
            field.setAccessible(true);
            fieldObject = field.get(null);
            
        } catch (Throwable e)
        {
            LOGGER.log(Level.CONFIG,e.getMessage(),e);
        }
        return fieldObject;
    }

    /**
     * Define the value of a static field of the Framework
     * 
     * @param fieldName
     *            the name of the field
     * @param fieldObject
     *            the value to set
     */
    private void setFrameworkFieldValue(String fieldName,
            Object fieldObject)
    {
        try
        {
            Field field = Framework.class.getDeclaredField(fieldName);
            field.setAccessible(true);
            field.set(null, fieldObject);
            
        } catch (Throwable e)
        {
            LOGGER.log(Level.CONFIG,e.getMessage(),e);  
        } 
    }

    /**
     * Define the packages list exported by the system
     * 
     * @param systemPackages
     *            comma separated list of packages names
     */
    private void setSystemPackagesFieldValue(String systemPackages)
    {
        if(systemPackages == null || systemPackages.length()==0)
        {
            return;
        }
        try
        {
            Class<?> bundleClassLoaderClass = Concierge.class.getClassLoader().loadClass(
                    "ch.ethz.iks.concierge.framework.BundleClassLoader");
            
            Field field = bundleClassLoaderClass.getDeclaredField("FRAMEWORK_PACKAGES");
            field.setAccessible(true);
            HashSet packagesSet = (HashSet) field.get(null);
            
            String[] packages = systemPackages.split(",");
            for(String pakage : packages)
            {  
                packagesSet.add(pakage);
            }
            
        } catch (Throwable e)
        {
            LOGGER.log(Level.CONFIG,e.getMessage(),e);  
        } 
    }
    
    /**
     * Invoke a static method of the Framework
     * 
     * @param methodName
     *            the name of the method
     * @param argsType
     *            A array of the method's parameters' classes
     * @param argsValue
     *            An array of objects used as parameters in the method
     */
    private Object invokeFrameworkMethod(String methodName,
            Class<?>[] argsType, Object[] argsValue)
    {
        try
        {
            Method method = Framework.class.getDeclaredMethod(methodName, 
                    argsType);
            method.setAccessible(true);
            return method.invoke(null, argsValue);
            
        } catch (Throwable e)
        {
            LOGGER.log(Level.CONFIG,e.getMessage(),e);  
        } 
        return null;
    }

    /**
     * Define the value of a field of the Framework's SystemBundle instance
     * 
     * @param fieldName
     *            the name of the field
     * @param fieldObject
     *            the value to set
     */
    private void setSystemBundleFieldValue(String fieldName,
            Object fieldObject)
    {
        if (systemBundle == null)
        {
            return;
        }
        try
        {
            Class<?> systemBundleClass = getFrameworkInnerClass(SYSTEMBUNDLE_CLASS_NAME);
            Field field = systemBundleClass.getDeclaredField(fieldName);
            field.setAccessible(true);
            field.set(systemBundle, fieldObject);
            
        }  catch (Throwable e)
        {
            LOGGER.log(Level.CONFIG,e.getMessage(),e);  
        } 
    }

    /**
     * Invoke a method of the Framework's SystemBundle instance
     * 
     * @param methodName
     *            the name of the method
     * @param argsType
     *            A array of the method's parameters' classes
     * @param argsValue
     *            An array of objects used as parameters by the method
     */
    private void invokeSystemBundleMethod(String methodName,
            Class<?>[] argsType, Object[] argsValue)
    {
        if (systemBundle == null)
        {
            return;
        }
        try
        {
            Class<?> systemBundleClass = getFrameworkInnerClass(SYSTEMBUNDLE_CLASS_NAME);
            Method method = systemBundleClass.getDeclaredMethod(methodName, argsType);
            method.setAccessible(true);
            method.invoke(systemBundle, argsValue);
            
        } catch (Throwable e)
        {
            LOGGER.log(Level.CONFIG,e.getMessage(),e);  
        } 
    }
    
    /**
     * ${inherited}
     * 
     * @see org.objectweb.fractal.juliac.osgi.Concierge#getBundleContext()
     */
    public BundleContext getBundleContext()
    {
        BundleContext context = null;
        if (systemBundle != null && bundleContextConstructor != null)
        {
            try
            {
                context = (BundleContext) 
                bundleContextConstructor.newInstance();
                
            } catch (Throwable e)
            {
                LOGGER.log(Level.CONFIG,e.getMessage(),e);  
            } 
        }
        return context;
    }

    /**
     * ${inherited}
     * 
     * @see org.objectweb.fractal.juliac.osgi.Concierge#stop()
     */
    public void stop()
    {
        if (systemBundle != null)
        {
            try
            {
                systemBundle.stop();
                
            } catch (BundleException e)
            {
                LOGGER.log(Level.CONFIG,e.getMessage(),e);
            }
        }
    }

    /**
     * Return true if the bundle are uncompressed before being stored
     * false otherwise
     */
    public boolean decompressed()
    {
        return ((Boolean) getFrameworkFieldValue(
                "DECOMPRESS_EMBEDDED")).booleanValue();
    }
    
    /**
     * Return the path of the Framework's storage location
     */
    public String storage()
    {
        return (String) getFrameworkFieldValue(
                "STORAGE_LOCATION");
    }

    /**
     * @param className
     * @return
     */
    public ClassLoader getBundleClassLoader(String packageName)
    {        
        Map exportedPackages = (Map)this.getFrameworkFieldValue("exportedPackages");

        Object[] exportedPackagesArray = exportedPackages.keySet().toArray(
                new Object[exportedPackages.size()]);    
       
        Class<?> packageClass = exportedPackagesArray[0].getClass();
        Object exportedPackage = null;    
        try
        {
            Method matches =  packageClass.getDeclaredMethod("matches",
                    new Class<?>[]{String.class});
            matches.setAccessible(true);
            
            for(Object exportedPackageItem : exportedPackagesArray)
            {
               if(((Boolean)matches.invoke(exportedPackageItem,
                       new Object[]{packageName})).booleanValue())
               {
                   exportedPackage = exportedPackageItem;
                   break;
               }
            }
            if(exportedPackage != null)
            {
                Field classloader = packageClass.getDeclaredField("classloader");
                classloader.setAccessible(true);
                return (ClassLoader) classloader.get(exportedPackage);
            }
        } catch(Throwable throwable)
        {
            if(LOGGER.isLoggable(Level.WARNING))
            {
                LOGGER.log(Level.WARNING,throwable.getMessage(),throwable);
            }
        }
        return null;
    }
    
    public Enumeration<Bundle> bundles()
    {
        final List bundles = (List) this.getFrameworkFieldValue("bundles");
        final int size = bundles.size();
        return new Enumeration<Bundle>()
        {
            int index = 0;
        
            public boolean hasMoreElements()
            {
                return index<size;
            }

            public Bundle nextElement()
            {
                return (Bundle)bundles.get(index++);
            }            
        };
    }
    
    /**
     * @param bundleId
     * @return
     */
    public ClassLoader getBundleClassLoader(long bundleId)
    {
        List bundles = (List) this.getFrameworkFieldValue("bundles");
        for(Object bundle : bundles)
        {
            if(((Bundle)bundle).getBundleId() == bundleId)
            {
                return getBundleClassLoader(((Bundle)bundle));
            }
        }
        return null;
    }
    
    /**
     * @param bundle
     * @return
     */
    public ClassLoader getBundleClassLoader(Bundle bundle)
    {
        ClassLoader classloader = null;
        Field classLoaderField = null;
        
        if(bundle.getBundleId() == 0)
        {
            return null;
        }
        if (bundleClass == null)
        {
            bundleClass = bundle.getClass();
        }
        try
        {
            classLoaderField = bundleClass.getDeclaredField("classloader");
            classLoaderField.setAccessible(true);
            classloader = (ClassLoader) classLoaderField.get(bundle);
            
            return classloader;
            
        } catch (Throwable e)
        {
            LOGGER.log(Level.CONFIG,e.getMessage(),e);  
        } 
        return null;
    }

    /**
     * @param bundleId
     * @return
     */
    public JarFile getBundleJar(long bundleId)
    {
        JarFile jarFile = null;
        ClassLoader classloader = getBundleClassLoader(getBundleContext().getBundle(bundleId));        
        Field jarFileField = null;
        try
        {
            jarFileField = classloader.getClass().getDeclaredField("jarFile");
            jarFileField.setAccessible(true);
            jarFile = (JarFile) jarFileField.get(classloader);
            return jarFile;
            
        } catch (Throwable e)
        {
            LOGGER.log(Level.CONFIG,e.getMessage(),e);  
        } 
        return null;
    }
}
